;; This file is part of Monocypher for CHICKEN
;; Copyright (c) 2020, Thomas Chust.  All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;;
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the
;;    distribution.
;;
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;; A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;; HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(require-library srfi-1 srfi-42 optimism stty iup protobuf monocypher)
(include-relative "protobuf.scm")

(import
  srfi-1
  srfi-4
  srfi-42
  optimism
  stty iup
  monocypher
  (monocypher protobuf)
  (chicken fixnum)
  (chicken process-context)
  (only (chicken format) printf)
  (only (chicken io) read-line read-string)
  (only (chicken port) port-for-each copy-port)
  (only (chicken gc) set-finalizer!)
  (only (chicken blob) blob-size string->blob)
  (only (chicken time posix) seconds->string))

(define key-magic "\x89M3C5k\r\n")
(define envelope-magic "\x89M3C5e\r\n")

(define (next-command opts command)
  (define commands
    (delete command '(help generate info hash sign verify sign-key verify-key box unbox stream version)))
  (let loop ([opts opts])
    (and
      (pair? opts)
      (let ([command (caar opts)])
        (if (memq command commands)
          command
          (loop (cdr opts)))))))

(define (get-password opts prompt)
  (set-finalizer!
    (if (assq '--gui opts)
      (letrec* ([close (lambda args (hide dlg))]
                [pwd (textbox #f #:password #t #:expand 'horizontal)]
                [ok (button "Ok" #:action close)]
                [dlg
                 (dialog
                   (vbox
                     (label prompt)
                     pwd
                     (hbox (fill) ok)
                     #:margin "5x5" #:gap 3)
                   #:title (program-name)
                   #:defaultenter ok)])
        (show dlg #:modal? #t)
        (attribute pwd 'value))
      (with-output-to-file "/dev/tty"
        (lambda ()
          (printf "~a: ~!" prompt)
          (let ([pwd
                 (with-input-from-file "/dev/tty"
                   (lambda ()
                     (with-stty '(not echo) read-line)))])
            (newline)
            pwd))))
    wipe!))

(define (get-protection-key opts)
  (derive-symmetric-box-key
    (get-password opts "Key protection password")
    (blob->u8vector/shared (string->blob key-magic))))

(define (get-signature-key opts)
  (derive-symmetric-sign-key
    (get-password opts "Signature password")
    (blob->u8vector/shared (string->blob envelope-magic))))

(define (get-encryption-key opts)
  (derive-symmetric-box-key
    (get-password opts "Encryption password")
    (blob->u8vector/shared (string->blob envelope-magic))))

(define (get-stream-key opts)
  (derive-random-stream-key
    (get-password opts "Stream password")
    (blob->u8vector/shared (string->blob envelope-magic))))

(define (get-public-key opts location)
  (cond
    [(assq '--public opts)
     => (lambda (opt)
          (call-with-input-file (cadr opt)
            (lambda (port)
              (unless (equal? (read-string 8 port) key-magic)
                (error location "invalid public key"))
              (read-key port))))]
    [else
     (error location "public key required")]))

(define (get-secret-key opts location)
  (cond
    [(assq '--secret opts)
     => (lambda (opt)
          (call-with-input-file (cadr opt)
            (lambda (port)
              (unless (equal? (read-string 8 port) key-magic)
                (error location "invalid secret key"))
              (let ([k (read-key port)])
                (if (boxed-key? k)
                  (or
                    ((symmetric-key-unbox (get-protection-key opts)) k)
                    (error location "invalid key protection password"))
                  k)))))]
    [else
     (error location "public key required")]))

(define (redirect-i/o opts #!key [i-keys '()] [o-keys '()])
  (let loop ([i-file "-"] [o-file "-"] [opts opts])
    (if (pair? opts)
      (let ([key (caar opts)]
            [val (cdar opts)])
        (cond
          [(and (memq key i-keys) (pair? val))
           (loop (car val) o-file (cdr opts))]
          [(and (memq key o-keys) (pair? val))
           (loop i-file (car val) (cdr opts))]
          [else
           (loop i-file o-file (cdr opts))]))
      (let ([i-port
             (and
               (not (string=? i-file "-"))
               (current-input-port (open-input-file i-file)))]
            [o-port
             (and
               (not (string=? o-file "-"))
               (current-output-port (open-output-file o-file)))])
        (or i-port o-port)))))

(define write-speak64
  (let ([alphabet
         #( "a"  "e"
           "ba" "be" "bi" "bo" "bu"
           "fa" "fe" "fi" "fo" "fu"
           "ha" "he" "hi" "ho" "hu"
           "ka" "ke" "ki" "ko" "ku"
           "ma" "me" "mi" "mo" "mu"
           "na" "ne" "ni" "no" "nu"
           "pa" "pe" "pi" "po" "pu"
           "ra" "re" "ri" "ro" "ru"
           "sa" "se" "si" "so" "su"
           "ta" "te" "ti" "to" "tu"
           "wa" "we" "wi" "wo"
           "ya"           "yo" "yu"
           "za" "ze" "zi" "zo" "zu")])
    (lambda (data #!optional [port (current-output-port)])
      (do ([i 0 (fx+ i 3)]) [(fx>= i (u8vector-length data))]
        (let* ([d (fxmin (fx- (u8vector-length data) i) 3)]
               [v (fold-ec 0 (:range i i (fx+ i d))
                    (u8vector-ref data i)
                    (lambda (v acc) (fxior (fxshl acc 8) v)))])
          (define (display-regular)
            (do-ec (:let w (fxshl v (fx* 8 (fx- 3 d)))) (:range j d)
              (display (vector-ref alphabet (fxand (fxshr w (fx* 6 (fx- 3 j))) #b111111)) port)))
          (if (fx= d 3)
            (cond
              [(fx= v #x000000) (display #\o port)]
              [(fx= v #xffffff) (display #\i port)]
              [else             (display-regular)])
            (display-regular)))))))

(define (do-help opts)
  (printf
    (case (next-command opts 'help)
      [(#f)
       #<<EOF
SYNOPSIS
    ~a COMMAND [OPTIONS...] - CLI for MonoCypher operations

COMMAND
    help [COMMAND]          - show information about the program or a command
    generate COMMAND        - generate a key for use with the given command
    info [INPUT]            - show information about data
    hash [INPUT]            - compute a message digest of data
    sign [INPUT]            - sign and authenticate data
    verify [INPUT]          - verify and authenticate data
    sign-key [INPUT]        - sign and authenticate a key
    verify-key [INPUT]      - verify and authenticate a key
    box [INPUT]             - encrypt and authenticate data
    unbox [INPUT]           - decrypt and authenticate data
    stream [INPUT]          - generate a pseudo-random stream [xored with data]
    version                 - show the version number of the program

OPTIONS
    --help, -h              - show information about the program or a command
    --version               - show the version number of the program
    --gui                   - use graphical user interface dialogs

EOF
       ]
      [(generate)
       #<<EOF
SYNOPSIS
    ~a generate COMMAND     - generate a key for use with the given command

OPTIONS
    --public FILE           - write the public key to a file
    --secret FILE, -o FILE  - write the secret key to a file
    --password              - encrypt the secret key with a password

EOF
       ]
      [(info)
       #<<EOF
SYNOPSIS
    ~a info [INPUT]         - show information about data

EOF
       ]
      [(hash)
       #<<EOF
SYNOPSIS
    ~a hash [INPUT]         - compute a message digest of data

EOF
       ]
      [(sign)
       #<<EOF
SYNOPSIS
    ~a sign [INPUT]         - sign and authenticate data

OPTIONS
    --secret FILE           - read the secret key from a file
    --password              - derive the secret key from a password
    -o FILE                 - write output to a file

EOF
       ]
      [(verify)
       #<<EOF
SYNOPSIS
    ~a verify [INPUT]       - verify and authenticate data

OPTIONS
    --public FILE           - read the public key from a file
    --password              - derive the secret key from a password
    -o FILE                 - write output to a file

EOF
       ]
      [(sign-key)
       #<<EOF
SYNOPSIS
    ~a sign-key [INPUT]     - sign and authenticate a key

OPTIONS
    --secret FILE           - read the secret key from a file
    --password              - derive the secret key from a password
    -o FILE                 - write output to a file

EOF
       ]
      [(verify-key)
       #<<EOF
SYNOPSIS
    ~a verify-key [INPUT]   - verify and authenticate a key

OPTIONS
    --public FILE           - read the public key from a file
    --password              - derive the secret key from a password

EOF
       ]
      [(box)
       #<<EOF
SYNOPSIS
    ~a box [INPUT]          - encrypt and authenticate data

OPTIONS
    --public FILE           - read the recipient's public key from a file
    --secret FILE           - read the sender's secret key from a file
    --password              - derive the secret key from a password
    -o FILE                 - write output to a file

EOF
       ]
      [(unbox)
       #<<EOF
SYNOPSIS
    ~a unbox [INPUT]        - decrypt and authenticate data

OPTIONS
    --public FILE           - read the recipient's public key from a file
    --secret FILE           - read the sender's secret key from a file
    --password              - derive the secret key from a password
    -o FILE                 - write output to a file

EOF
       ]
      [(stream)
       #<<EOF
SYNOPSIS
    ~a stream [INPUT]       - generate a pseudo-random stream [xored with data]

OPTIONS
    --public FILE           - read the recipient's public key from a file
    --secret FILE           - read the sender's secret key from a file
    --password              - derive the secret key from a password
    --limit NUMBER          - limit the size of the output
    -o FILE                 - write output to a file

EOF
       ]
      [(version --version)
       #<<EOF
SYNOPSIS
    ~a version              - show the version number of the program

EOF
       ]
      [else
       => (lambda (command) (error 'help "unknown command" command))])
    (program-name)))

(define (do-generate opts)
  (define generate-key
    (case (next-command opts 'generate)
      [(sign verify sign-key verify-key)
       generate-asymmetric-sign-key]
      [(box unbox stream)
       generate-asymmetric-box-key]
      [(#f)
       (error 'generate "missing command argument")]
      [else
       => (lambda (command) (error 'generate "command uses no asymmetric keys" command))]))
  (redirect-i/o opts #:o-keys '(--secret -o))
  (let ([k (generate-key)])
    (when (assq '--password opts)
      (set! k ((symmetric-key-box (get-protection-key opts)) k)))
    (display key-magic)
    (write-key k)
    (when (redirect-i/o opts #:o-keys '(--public))
      (display key-magic)
      (write-key (make-public-key k)))))

(define (do-info opts)
  (redirect-i/o opts #:i-keys '(--))
  (let ([magic (read-string 8)])
    (cond
      [(eof-object? magic)
       (printf "Empty~%")]
      [(string=? magic key-magic)
       (printf "Key~%")
       (let ([k (read-key)])
         (printf "    algorithm: ~a~%" (key-algorithm k))
         (cond
           [(public-key k)
            => (lambda (pk)
                 (printf "    public:    ~a bit" (fx* (blob-size pk) 8))
                 (cond
                   [(public-key-envelope k)
                    => (lambda (e)
                         (printf ", ~a" (seconds->string (round (nonce-seconds (envelope-nonce e))))))])
                 (printf "~%               ")
                 (write-speak64 (blob->u8vector/shared pk))
                 (newline))])
         (cond
           [(secret-key k)
            => (lambda (sk)
                 (if (envelope? sk)
                   (printf "    secret:    encrypted~%")
                   (printf "    secret:    ~a bit~%" (fx* (blob-size sk) 8))))]))]
      [(string=? magic envelope-magic)
       (printf "Envelope~%")
       (let ([e (read-envelope)])
         (printf "    algorithm: ~a~%" (envelope-algorithm e))
         (cond
           [(envelope-nonce e)
            => (lambda (n)
                 (printf "    nonce:     ~a byte, ~a~%" (u8vector-length n) (seconds->string (round (nonce-seconds n)))))])
         (printf "    message:   ~a byte~%" (string-length (envelope-message e)))
         (cond
           [(envelope-signature e)
            => (lambda (s)
                 (printf "    signature: ~a bit~%" (fx* (blob-size s) 8)))]))]
      [else
       (printf "Data~%")])))

(define (do-hash opts)
  (redirect-i/o opts #:i-keys '(--))
  (write-speak64 (blob->u8vector/shared (hash (read-string))))
  (newline))

(define (do-sign opts)
  (define sign
    (if (assq '--password opts)
      (symmetric-sign (get-signature-key opts))
      (asymmetric-sign (get-secret-key opts 'sign))))
  (redirect-i/o opts #:i-keys '(--) #:o-keys '(-o))
  (display envelope-magic)
  (write-envelope (sign (read-string))))

(define (do-verify opts)
  (define verify
    (if (assq '--password opts)
      (symmetric-verify (get-signature-key opts))
      (asymmetric-verify (get-public-key opts 'sign))))
  (redirect-i/o opts #:i-keys '(--) #:o-keys '(-o))
  (unless (equal? (read-string 8) envelope-magic)
    (error 'verify "invalid envelope"))
  (cond
    [(verify (read-envelope)) => display]
    [else (error 'verify "invalid envelope or signature key")]))

(define (do-sign-key opts)
  (define sign
    (if (assq '--password opts)
      (symmetric-key-sign (get-signature-key opts))
      (asymmetric-key-sign (get-secret-key opts 'sign-key))))
  (redirect-i/o opts #:i-keys '(--) #:o-keys '(-o))
  (unless (equal? (read-string 8) key-magic)
    (error 'sign-key "invalid key"))
  (display key-magic)
  (write-key (sign (read-key))))

(define (do-verify-key opts)
  (define verify
    (if (assq '--password opts)
      (symmetric-key-verify (get-signature-key opts))
      (asymmetric-key-verify (get-public-key opts 'sign-key))))
  (redirect-i/o opts #:i-keys '(--))
  (unless (equal? (read-string 8) key-magic)
    (error 'verify-key "invalid key"))
  (unless (verify (read-key))
    (error 'verify-key "invalid key or signature key")))

(define (do-box opts)
  (define box
    (symmetric-box
      (if (assq '--password opts)
        (get-encryption-key opts)
        (exchange-symmetric-box-key
          (get-public-key opts 'box)
          (get-secret-key opts 'box)))))
  (redirect-i/o opts #:i-keys '(--) #:o-keys '(-o))
  (display envelope-magic)
  (write-envelope (box (read-string))))

(define (do-unbox opts)
  (define unbox
    (symmetric-unbox
      (if (assq '--password opts)
        (get-encryption-key opts)
        (exchange-symmetric-box-key
          (get-public-key opts 'unbox)
          (get-secret-key opts 'unbox)))))
  (redirect-i/o opts #:i-keys '(--) #:o-keys '(-o))
  (unless (equal? (read-string 8) envelope-magic)
    (error 'unbox "invalid envelope"))
  (cond
    [(unbox (read-envelope)) => display]
    [else (error 'unbox "invalid envelope or encryption key")]))

(define (do-stream opts)
  (define stream
    (open-random-stream
      (if (assq '--password opts)
        (get-stream-key opts)
        (exchange-random-stream-key
          (get-public-key opts 'stream)
          (get-secret-key opts 'stream)))
      #:limit
      (cond
        [(assq '--limit opts) => cdr]
        [else #f])
      #:nonce
      (make-u8vector random-stream-noncebytes 0)))
  (redirect-i/o opts #:o-keys '(-o))
  (if (redirect-i/o opts #:i-keys '(--))
    (port-for-each
      (lambda (buf)
        (stream-xor! buf stream)
        (display buf))
      (lambda ()
        (read-string 4096)))
    (copy-port stream (current-output-port)))
  (close-input-port stream))

(define (do-version opts)
  (printf "~a 1.0.0~%" (program-name)))

(let ([opts
       (parse-command-line
         `([help] [--help] [-h]
           [generate] [--password] [--gui]
           [info]
           [hash]
           [sign] [--secret file]
           [verify]
           [sign-key]
           [verify-key]
           [box] [--public file]
           [unbox]
           [stream] [--limit . ,string->number]
           [version] [--version]
           [-o file]))])
  (cond
    [(or (assq 'help opts) (assq '--help opts) (assq '-h opts))
     (do-help opts)]
    [(assq 'generate opts)
     (do-generate opts)]
    [(assq 'info opts)
     (do-info opts)]
    [(assq 'hash opts)
     (do-hash opts)]
    [(assq 'sign opts)
     (do-sign opts)]
    [(assq 'verify opts)
     (do-verify opts)]
    [(assq 'sign-key opts)
     (do-sign-key opts)]
    [(assq 'verify-key opts)
     (do-verify-key opts)]
    [(assq 'box opts)
     (do-box opts)]
    [(assq 'unbox opts)
     (do-unbox opts)]
    [(assq 'stream opts)
     (do-stream opts)]
    [(or (assq 'version opts) (assq '--version opts))
     (do-version opts)]
    [else
     (do-help opts)
     (exit 1)]))

;; vim: set ft=scheme sts=2 sw=2 ts=4 et ai: ;;
