;; This file is part of Monocypher for CHICKEN
;; Copyright (c) 2020, Thomas Chust.  All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;;
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the
;;    distribution.
;;
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;; A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;; HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(import
  (only (chicken blob) string->blob)
  monocypher srfi-4 test)

(define plain "Hello world! Let's use Monocypher for private communication :-)")

(test-group "asymmetric algorithms"

  (test-group "boxes"
    (let* ([alice-sk (make-key "X25519/HChaCha20" #${38ec1d223d95b4a71ba93d1dec4aa99dadc9887f10eb467ab0e148bea0bea03d} #${6c0d4a72b5f7e7f816f5f0f9ab992893a10f49b8d61eaa3c2d3ddec6c76efa24})]
           [alice-pk (make-public-key alice-sk)]
           [bob-sk (make-key "X25519/HChaCha20" #${1a29d867ee740c0c6acec12e21a5b4f845d237e5236e3769ec3ce4fe61822b07} #${284e1c6e7b244d432d56b5d210be997ad703d1f4aae873efab1e3ad961a40ee9})]
           [bob-pk (make-public-key bob-sk)]
           [bad-sk (make-key "X25519/HChaCha20" #f #${284e1c6e7b244d43fffffffffffffffffffffffffffff3efab1e3ad961a40ee9})]
           [nonce (make-u8vector asymmetric-box-noncebytes 42)]
           [bad-nonce (make-u8vector asymmetric-box-noncebytes 23)]
           [crypt ((asymmetric-box bob-pk alice-sk) plain #:nonce nonce)])
      (test "box roundtrip" plain ((asymmetric-unbox alice-pk bob-sk) crypt))
      (test "secret box compatibility" plain ((symmetric-unbox (exchange-symmetric-box-key alice-pk bob-sk)) crypt))
      (test-assert "bad nonce detection" (not ((asymmetric-unbox alice-pk bob-sk) (make-envelope (envelope-algorithm crypt) bad-nonce (envelope-message crypt) (envelope-signature crypt)))))
      (test-assert "bad secret key detection" (not ((asymmetric-unbox alice-pk bad-sk) crypt)))
      (string-set! (envelope-message crypt) 23 #\nul)
      (test-assert "damage detection" (not ((asymmetric-unbox alice-pk bob-sk) crypt)))))

  (test-group "boxes with hidden public keys"
    (let* ([alice-sk (make-key "X25519/HChaCha20" #${736ce148b60da0d3eb5d3bc87ec35de71bc83d2b52c4b9e2ab94a6d07b1c514d} #${6c579709327d1eda2d92287480d5ceb6287ecd48ac75aad6c48fdd638d26b39f})]
	   [alice-hk (hide-asymmetric-box-key alice-sk #:tweak 23)]
	   [bob-sk (make-key "X25519/HChaCha20" #${6ca030e5659f30aaa0209941abf402bae446f274242366c4a085de4d34475414} #${dec76a894c2ec03ff81323ae974c5772771bd0733c8066ecfb94ef650171fbf9})]
	   [bob-hk (hide-asymmetric-box-key bob-sk #:tweak 42)]
	   [bad-sk (make-key "X25519/HChaCha20" #f #${dec76a894c2ec03ffffffffffffffffff71bd0733c8066ecfb94ef650171fbf9})]
           [nonce (make-u8vector asymmetric-box-noncebytes 66)]
           [bad-nonce (make-u8vector asymmetric-box-noncebytes 99)]
	   [crypt ((asymmetric-box (reveal-asymmetric-box-key bob-hk) alice-sk) plain #:nonce nonce)])
      (test "box roundtrip" plain ((asymmetric-unbox (reveal-asymmetric-box-key alice-hk) bob-sk) crypt))
      (test "secret box compatibility" plain ((symmetric-unbox (exchange-symmetric-box-key (reveal-asymmetric-box-key alice-hk) bob-sk)) crypt))
      (test-assert "bad nonce detection" (not ((asymmetric-unbox (reveal-asymmetric-box-key alice-hk) bob-sk) (make-envelope (envelope-algorithm crypt) bad-nonce (envelope-message crypt) (envelope-signature crypt)))))
      (test-assert "bad secret key detection" (not ((asymmetric-unbox (reveal-asymmetric-box-key alice-hk) bad-sk) crypt)))
      (string-set! (envelope-message crypt) 23 #\nul)
      (test-assert "damage detection" (not ((asymmetric-unbox (reveal-asymmetric-box-key alice-hk) bob-sk) crypt)))))

  (test-group "signatures"
    (let* ([eve-sk (make-key "EdDSA/Curve25519" #${0046eda05901fd08fba4286c71cc325badfd912bb837d5f74820e4541a9ada52} #${3760b5a82a79bc9973a26e7b0a236b4e1f439deae59ae978d0ab66e58cb2229a0046eda05901fd08fba4286c71cc325badfd912bb837d5f74820e4541a9ada52})]
           [eve-pk (make-public-key eve-sk)]
           [bad-pk (make-key "EdDSA/Curve25519" #${0046eda05901fd08fba4286c71fffffffffd912bb837d5f74820e4541a9ada52} #f)]
           [signed ((asymmetric-sign eve-sk) (string-copy plain) #:nonce #f)])
      (test "sign roundtrip" plain ((asymmetric-verify eve-pk) signed))
      (test-assert "bad public key detection" (not ((asymmetric-verify bad-pk) signed)))
      (string-set! (envelope-message signed) 23 #\nul)
      (test-assert "damage detection" (not ((asymmetric-verify eve-pk) signed)))))

)

(test-group "symmetric algorithms"

  (test-group "boxes"
    (let* ([eve-k (make-key "XChaCha20/Poly1305" #f #${b4d3f64119e17ab5da5bb5618449e690fed752594bd72277e6f20caee12d1a0c})]
           [bad-k (make-key "XChaCha20/Poly1305" #f #${b4d3f64ffffffffffffff5618449e690ffffffffffffff77e6f20caee12d1a0c})]
           [nonce (make-u8vector symmetric-box-noncebytes 42)]
           [bad-nonce (make-u8vector symmetric-box-noncebytes 23)]
           [crypt ((symmetric-box eve-k) plain #:nonce nonce)])
      (test "box roundtrip" plain ((symmetric-unbox eve-k) crypt))
      (test-assert "bad nonce detection" (not ((symmetric-unbox eve-k) (make-envelope (envelope-algorithm crypt) bad-nonce (envelope-message crypt) (envelope-signature crypt)))))
      (test-assert "bad key detection" (not ((symmetric-unbox bad-k) crypt)))
      (string-set! (envelope-message crypt) 23 #\nul)
      (test-assert "damage detection" (not ((symmetric-unbox eve-k) crypt)))))

  (test-group "one-time signatures"
    (let* ([eve-k (make-key "Poly1305" #f #${7abfa988ef67acfd36ae0ecfebc7997d377e6012df62ad6ff53d2030cc33596b})]
           [bad-k (make-key "Poly1305" #f #${7abfa988ef67acfd36ae0ecfebc7997d377e601ffffffffffffffffffff3596b})]
           [signed ((symmetric-sign eve-k) (string-copy plain) #:nonce #f)])
      (test "sign roundtrip" plain ((symmetric-verify eve-k) signed))
      (test-assert "bad key detection" (not ((symmetric-verify bad-k) signed)))
      (string-set! (envelope-message signed) 23 #\nul)
      (test-assert "damage detection" (not ((symmetric-verify eve-k) signed)))))

  (test-group "random streams"
    (let* ([alice-sk (make-key "X25519/HChaCha20" #${38ec1d223d95b4a71ba93d1dec4aa99dadc9887f10eb467ab0e148bea0bea03d} #${6c0d4a72b5f7e7f816f5f0f9ab992893a10f49b8d61eaa3c2d3ddec6c76efa24})]
           [alice-pk (make-public-key alice-sk)]
           [bob-sk (make-key "X25519/HChaCha20" #${1a29d867ee740c0c6acec12e21a5b4f845d237e5236e3769ec3ce4fe61822b07} #${284e1c6e7b244d432d56b5d210be997ad703d1f4aae873efab1e3ad961a40ee9})]
           [bob-pk (make-public-key bob-sk)]
           [rng-k (make-key "XChaCha20" #f #${4b94b23477e98dd33f657c9f657251bc9a93796f3b7176697e1744674d46aea6})]
           [nonce (make-u8vector random-stream-noncebytes 42)])
      (test "xor roundtrip" plain (stream-xor! (stream-xor plain (open-random-stream rng-k #:nonce nonce)) (open-random-stream rng-k #:nonce nonce)))
      (test "key agreement" plain (stream-xor! (stream-xor plain (open-random-stream (exchange-random-stream-key alice-pk bob-sk) #:nonce nonce)) (open-random-stream (exchange-random-stream-key bob-pk alice-sk) #:nonce nonce)))))

  (test "hash"
    #${c2edf2ad1e46a8da2ad212161d775830a4123c37b210fa5779a03150224250bd1b4584eb37f37f420c2c87b5d324afb39d5d9a89f437558816036eaf3323d82f}
    (hash plain))

)

(test-exit)
;; vim: set ft=scheme sts=2 sw=2 ts=4 et ai: ;;
